# sage_setup: distribution = sagemath-graphs
# cython: binding=True
r"""
Asteroidal triples

This module contains the following function:

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`is_asteroidal_triple_free` | Test if the input graph is asteroidal triple-free

Definition
----------

Three independent vertices of a graph form an *asteroidal triple* if every two
of them are connected by a path avoiding the neighborhood of the third one. A
graph is *asteroidal triple-free* (*AT-free*, for short) if it contains no
asteroidal triple [LB1962]_.

Use ``graph_classes.AT_free.description()`` to get some known properties of
AT-free graphs, or visit `this page
<http://www.graphclasses.org/classes/gc_61.html>`_.


Algorithm
---------

This module implements the  *Straightforward algorithm* recalled in [Koh2004]_ and
due to [LB1962]_ for testing if a graph is AT-free or not. This algorithm has time
complexity in `O(n^3)` and space complexity in `O(n^2)`.

This algorithm uses the *connected structure* of the graph, stored into a
`n\times n` matrix `M`. This matrix is such that `M[u][v]==0` if `v\in
(\{u\}\cup N(u))`, and otherwise `M[u][v]` is the unique identifier (a strictly
positive integer) of the connected component of `G\setminus(\{u\}\cup N(u))` to
which `v` belongs. This connected structure can be computed in time `O(n(n+m))`
using `n` BFS.

Now, a triple `u, v, w\in V` is an asteroidal triple if and only if it satisfies
`M[u][v]==M[u][w]` and `M[v][u]==M[v][w]` and `M[w][u]==M[w][v]`, assuming all
these values are positive. Indeed, if `M[u][v]==M[u][w]`, `v` and `w` are in the
same connected component of `G\setminus(\{u\}\cup N(u))`, and so there is a path
between `v` and `w` avoiding the neighborhood of `u`. The algorithm iterates
over all triples.


Functions
---------
"""

# ****************************************************************************
#       Copyright (C) 2015 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from libc.stdint cimport uint32_t
from cysignals.signals cimport sig_on, sig_off
from memory_allocator cimport MemoryAllocator

from sage.data_structures.bitset_base cimport *
from sage.graphs.base.static_sparse_graph cimport short_digraph, init_short_digraph, free_short_digraph


def is_asteroidal_triple_free(G, certificate=False):
    """
    Test if the input graph is asteroidal triple-free.

    An independent set of three vertices such that each pair is joined by a path
    that avoids the neighborhood of the third one is called an *asteroidal
    triple*. A graph is asteroidal triple-free (AT-free) if it contains no
    asteroidal triples. See the :mod:`module's documentation
    <sage.graphs.asteroidal_triples>` for more details.

    This method returns ``True`` if the graph is AT-free and ``False`` otherwise.

    INPUT:

    - ``G`` -- a Graph

    - ``certificate`` -- boolean (default: ``False``); by default, this method
      returns ``True`` if the graph is asteroidal triple-free and ``False``
      otherwise. When ``certificate==True``, this method returns in addition a
      list of three vertices forming an asteroidal triple if such a triple is
      found, and the empty list otherwise.

    EXAMPLES:

    The complete graph is AT-free, as well as its line graph::

        sage: G = graphs.CompleteGraph(5)
        sage: G.is_asteroidal_triple_free()
        True
        sage: G.is_asteroidal_triple_free(certificate=True)
        (True, [])
        sage: LG = G.line_graph()
        sage: LG.is_asteroidal_triple_free()
        True
        sage: LLG = LG.line_graph()
        sage: LLG.is_asteroidal_triple_free()
        False

    The PetersenGraph is not AT-free::

        sage: from sage.graphs.asteroidal_triples import *
        sage: G = graphs.PetersenGraph()
        sage: G.is_asteroidal_triple_free()
        False
        sage: G.is_asteroidal_triple_free(certificate=True)
        (False, [0, 2, 6])

    TESTS:

    Giving anything else than a Graph::

        sage: from sage.graphs.asteroidal_triples import is_asteroidal_triple_free
        sage: is_asteroidal_triple_free(DiGraph())
        Traceback (most recent call last):
        ...
        ValueError: The first parameter must be a Graph.
    """
    from sage.graphs.graph import Graph
    if not isinstance(G, Graph):
        raise ValueError("The first parameter must be a Graph.")

    cdef uint32_t n = <uint32_t>G.order()
    cdef int i

    # ==> Trivial cases
    if n < 3:
        return (True, []) if certificate else True

    # ==> Initialize some data structures for is_asteroidal_triple_free_C
    cdef MemoryAllocator mem = MemoryAllocator()
    cdef uint32_t* waiting_list = <uint32_t*> mem.allocarray(n, sizeof(uint32_t))
    cdef uint32_t* _connected_structure = <uint32_t*> mem.calloc(n * n, sizeof(uint32_t))
    cdef uint32_t** connected_structure = <uint32_t**> mem.allocarray(n, sizeof(uint32_t*))

    # Copying the whole graph to obtain the list of neighbors quicker than by
    # calling out_neighbors. This data structure is well documented in the
    # module sage.graphs.base.static_sparse_graph
    cdef list int_to_vertex = list(G)
    cdef short_digraph sd
    init_short_digraph(sd, G, edge_labelled=False, vertex_list=int_to_vertex)

    cdef bitset_t seen
    bitset_init(seen, n)

    connected_structure[0] = _connected_structure
    for i in range(n-1):
        connected_structure[i+1] = connected_structure[i] + n

    cdef list ret = []

    # ==> call is_asteroidal_triple_free_C

    try:
        sig_on()
        ret = is_asteroidal_triple_free_C(n, sd, connected_structure, waiting_list, seen)
        sig_off()

    finally:
        # Release memory
        bitset_free(seen)
        free_short_digraph(sd)

    # ==> We return the result

    if certificate:
        if ret:
            return False, [int_to_vertex[i] for i in ret]
        return True, []

    return False if ret else True


cdef list is_asteroidal_triple_free_C(uint32_t n,
                                      short_digraph sd,
                                      uint32_t** connected_structure,
                                      uint32_t* waiting_list,
                                      bitset_t seen):
    """
    INPUT:

    - ``n`` -- integer; number of points in the graph

    - ``sd`` -- ``short_digraph``; a graph on ``n`` points. This data structure
      is well documented in the module
      :mod:`sage.graphs.base.static_sparse_graph`

    - ``connected_structure`` -- bidimensional array of size `n\times n` used to
      store the connected structure of the graph. All its cells must initially
      be set to 0.

    - ``waiting_list`` -- an array of size `n` to be used for BFS

    - ``seen`` -- a bitset of size `n`

    ALGORITHM:

    See the module's documentation.
    """
    cdef uint32_t waiting_beginning = 0
    cdef uint32_t waiting_end = 0
    cdef uint32_t idx_cc = 0
    cdef uint32_t source, u, v, w
    cdef uint32_t* p_tmp
    cdef uint32_t* end

    # ==> We build the connected structure

    # We run n different BFS taking each vertex as a source
    for source in range(n):

        # The source is forbidden and seen
        bitset_clear(seen)
        bitset_add(seen, source)

        # The neighbors of the source are forbidden and seen
        p_tmp = sd.neighbors[source]
        end = sd.neighbors[source + 1]
        # Iterating over all the outneighbors u of v
        while p_tmp < end:
            bitset_add(seen, p_tmp[0])
            p_tmp += 1

        # We now search for an unseen vertex
        v = bitset_first_in_complement(seen)
        while v != <uint32_t>-1:
            # and add it to the queue
            waiting_list[0] = v
            waiting_beginning = 0
            waiting_end = 0

            # We start a new connected component
            idx_cc += 1
            bitset_add(seen, v)
            connected_structure[source][v] = idx_cc

            # For as long as there are vertices left to explore in this
            # component
            while waiting_beginning <= waiting_end:

                # We pick the first one
                v = waiting_list[waiting_beginning]
                p_tmp = sd.neighbors[v]
                end = sd.neighbors[v + 1]

                # Iterating over all the outneighbors u of v
                while p_tmp < end:
                    u = p_tmp[0]

                    # If we notice one of these neighbors is not seen yet, we
                    # add it to the queue to be explored later
                    if not bitset_in(seen, u):
                        waiting_end += 1
                        waiting_list[waiting_end] = u
                        bitset_add(seen, u)
                        connected_structure[source][u] = idx_cc

                    p_tmp += 1

                waiting_beginning += 1

            # We search for a possibly unseen vertex
            v = bitset_first_in_complement(seen)

    # ==> Now that we have the component structure of the graph, we search for
    # an asteroidal triple.

    # (Possible improvement) right now, the code fixes u and tries to find v,w
    # in the same connected component of G-N[u] by going over all
    # binomial(n-1,2) pairs of point. It would be faster to:
    #
    # - Iterate on all connected components of G-N[u]
    # - Enumerate all v,w in G-N[u]
    #
    # The list of connected components of G-N[u] can be built from
    # connected_structure in O(n) time.

    for u in range(n - 2):
        for v in range(u + 1, n - 1):
            if connected_structure[u][v]:
                for w in range(v + 1, n):
                    if (connected_structure[u][v] == connected_structure[u][w] and
                            connected_structure[v][u] == connected_structure[v][w] and
                            connected_structure[w][u] == connected_structure[w][v]):
                        # We have found an asteroidal triple
                        return [u, v, w]

    # No asteroidal triple was found
    return []



# sage_setup: distribution = sagemath-graphs
"""
Mobile posets
"""
# ****************************************************************************
#       Copyright (C) 2020 Stefan Grosser <stefan.grosser1@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.posets.posets import Poset, FinitePoset
from sage.misc.lazy_attribute import lazy_attribute
from .linear_extensions import LinearExtensionsOfMobile


class MobilePoset(FinitePoset):
    r"""
    A mobile poset.

    Mobile posets are an extension of d-complete posets which permit a determinant
    formula for counting linear extensions. They are formed by having a ribbon
    poset with d-complete posets 'hanging' below it and at most one
    d-complete poset above it, known as the anchor. See [GGMM2020]_
    for the definition.

    EXAMPLES::

        sage: P = posets.MobilePoset(posets.RibbonPoset(7, [1,3]),                      # needs sage.combinat sage.modules
        ....:                        {1: [posets.YoungDiagramPoset([3, 2], dual=True)],
        ....:                         3: [posets.DoubleTailedDiamond(6)]},
        ....:                        anchor=(4, 2, posets.ChainPoset(6)))
        sage: len(P._ribbon)                                                            # needs sage.combinat sage.modules
        8
        sage: P._anchor                                                                 # needs sage.combinat sage.modules
        (4, 5)

    This example is Example 5.9 in [GGMM2020]_::

        sage: P1 = posets.MobilePoset(posets.RibbonPoset(8, [2,3,4]),
        ....:                         {4: [posets.ChainPoset(1)]},
        ....:                         anchor=(3, 0, posets.ChainPoset(1)))
        sage: sorted([P1._element_to_vertex(i) for i in P1._ribbon])
        [0, 1, 2, 6, 7, 9]
        sage: P1._anchor
        (3, 2)

        sage: P2 = posets.MobilePoset(posets.RibbonPoset(15, [1,3,5,7,9,11,13]),
        ....:                         {}, anchor=(8, 0, posets.ChainPoset(1)))
        sage: sorted(P2._ribbon)
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]
        sage: P2._anchor
        (8, (8, 0))
        sage: P2.linear_extensions().cardinality()                                      # needs sage.modules
        21399440939

        sage: EP = posets.MobilePoset(posets.ChainPoset(0), {})
        Traceback (most recent call last):
        ...
        ValueError: the empty poset is not a mobile poset
    """
    _lin_ext_type = LinearExtensionsOfMobile
    _desc = 'Finite mobile poset'

    def __init__(self, hasse_diagram, elements, category, facade, key, ribbon=None, check=True) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: P = posets.MobilePoset(posets.RibbonPoset(15, [1,3,5,7,9,11,13]),
            ....:                        {}, anchor=(8, 0, posets.ChainPoset(1)))
            sage: TestSuite(P).run()
        """
        FinitePoset.__init__(self, hasse_diagram=hasse_diagram, elements=elements,
                             category=category, facade=facade, key=key)
        if not self._hasse_diagram:
            raise ValueError("the empty poset is not a mobile poset")

        if ribbon:
            if check and not self._is_valid_ribbon(ribbon):
                raise ValueError("invalid ribbon")
            self._ribbon = ribbon

    def _is_valid_ribbon(self, ribbon):
        r"""
        Return ``True`` if a ribbon has at most one anchor, no vertex has two
        or more anchors, and every hanging poset is d-complete.

        INPUT:

        - ``ribbon`` -- list of elements that form a ribbon in your poset

        TESTS::

            sage: P = posets.RibbonPoset(5, [2])
            sage: P._is_valid_ribbon([0,1,2,3,4])
            True
            sage: P._is_valid_ribbon([2])
            False
            sage: P._is_valid_ribbon([2,3,4])
            True
            sage: P._is_valid_ribbon([2,3])
            True
        """
        ribbon = [self._element_to_vertex(x) for x in ribbon]
        G = self._hasse_diagram
        G_un = G.to_undirected().copy(immutable=False)
        R = G.subgraph(ribbon)
        num_anchors = 0

        for r in ribbon:
            anchor_neighbors = set(G.neighbor_out_iterator(r)).difference(set(R.neighbor_out_iterator(r)))
            if len(anchor_neighbors) == 1:
                num_anchors += 1
            elif len(anchor_neighbors) > 1:
                return False

            for lc in G.neighbor_in_iterator(r):
                if lc in ribbon:
                    continue

                G_un.delete_edge(lc, r)
                P = Poset(G.subgraph(G_un.connected_component_containing_vertex(lc, sort=False)))
                if P.top() != lc or not P.is_d_complete():
                    return False
                G_un.add_edge(lc, r)

        return True

    @lazy_attribute
    def _anchor(self):
        r"""
        The anchor of the mobile poset.

        TESTS::

            sage: from sage.combinat.posets.mobile import MobilePoset
            sage: M = MobilePoset(DiGraph([[0,1,2,3,4,5,6,7,8],
            ....:   [(1,0),(3,0),(2,1),(2,3),(4,3), (5,4),(5,6),(7,4),(7,8)]]))
            sage: M._anchor
            (4, 3)
        """
        ribbon = [self._element_to_vertex(x) for x in self._ribbon]
        H = self._hasse_diagram
        R = H.subgraph(ribbon)

        anchor = None

        # Find the anchor vertex, if it exists, and return the edge
        for r in ribbon:
            anchor_neighbors = set(H.neighbor_out_iterator(r)).difference(set(R.neighbor_out_iterator(r)))
            if len(anchor_neighbors) == 1:
                anchor = (r, anchor_neighbors.pop())
                break
        return (self._vertex_to_element(anchor[0]), self._vertex_to_element(anchor[1])) if anchor is not None else None

    @lazy_attribute
    def _ribbon(self):
        r"""
        The ribbon of the mobile poset.

        TESTS::

            sage: from sage.combinat.posets.mobile import MobilePoset
            sage: M = MobilePoset(DiGraph([[0,1,2,3,4,5,6,7,8],
            ....:   [(1,0),(3,0),(2,1),(2,3),(4,3), (5,4),(5,6),(7,4),(7,8)]]))
            sage: sorted(M._ribbon)
            [4, 5, 6, 7, 8]
            sage: M._is_valid_ribbon(M._ribbon)
            True
            sage: M2 = MobilePoset(Poset([[0,1,2,3,4,5,6,7,8],
            ....:          [(1,0),(3,0),(2,1),(2,3),(4,3),(5,4),(7,4),(7,8)]]))
            sage: sorted(M2._ribbon)
            [4, 7, 8]
            sage: M2._is_valid_ribbon(M2._ribbon)
            True
        """
        H = self._hasse_diagram
        H_un = H.to_undirected()
        max_elmts = H.sinks()
        # Compute anchor, ribbon
        ribbon = []  # In order list of elements on zigzag

        if len(max_elmts) == 1:
            return [self._vertex_to_element(max_elmts[0])]
        # Compute max element tree by merging shortest paths
        start = max_elmts[0]

        zigzag_elmts = set()
        for m in max_elmts[1:]:
            sp = H_un.shortest_path(start, m)
            zigzag_elmts.update(sp)
        max_elmt_graph = H.subgraph(zigzag_elmts)
        G = max_elmt_graph.to_undirected()

        if G.is_path():
            # Check if there is a anchor by seeing if there is more than one acyclic path to the next max
            ends = max_elmt_graph.vertices(sort=True, degree=1)
            # Form ribbon
            ribbon = G.shortest_path(ends[0], ends[1])
            for end_count, end in enumerate(ends):
                if not (H_un.is_cut_vertex(end) or H_un.degree(end) == 1):
                    traverse_ribbon = ribbon if end_count == 0 else ribbon[::-1]
                    for ind, p in enumerate(traverse_ribbon):
                        if H_un.is_cut_edge(p, traverse_ribbon[ind + 1]):
                            return [self._vertex_to_element(r)
                                    for r in G.shortest_path(ends[(end_count + 1) % 2], traverse_ribbon[ind + 1])]
            return [self._vertex_to_element(r) for r in ribbon]

        # First check path counts between ends and deg3 vertex
        # Then check if more than one max elmt on way to degree 3 vertex.
        # Then check if the edge going to a max element is down from the degree 3 vertex
        # Arbitrarily choose between ones with just 1

        ends = max_elmt_graph.vertices(sort=True, degree=1)
        deg3 = max_elmt_graph.vertices(sort=True, degree=3)[0]

        anchoredEnd = None
        for end in ends:
            if not (H_un.is_cut_vertex(end) or H_un.degree(end) == 1):
                anchoredEnd = end
                break

        if anchoredEnd is not None:
            ends.remove(anchoredEnd)
            return [self._vertex_to_element(r) for r in G.shortest_path(ends[0], ends[1])]

        possible_anchors = ends[:]
        for end in ends:
            path = G.shortest_path(end, deg3)
            if sum(bool(z in max_elmts) for z in path) != 1:
                possible_anchors.remove(end)

        for p in possible_anchors:
            path = G.shortest_path(p, deg3)
            if max_elmt_graph.has_edge(path[-2], path[-1]):
                possible_anchors.remove(p)

        anchoredEnd = possible_anchors[0]
        ends.remove(anchoredEnd)
        return [self._vertex_to_element(r) for r in G.shortest_path(ends[0], ends[1])]

    def ribbon(self):
        r"""
        Return the ribbon of the mobile poset.

        EXAMPLES::

            sage: from sage.combinat.posets.mobile import MobilePoset
            sage: M3 = MobilePoset(Posets.RibbonPoset(5, [1,2]))
            sage: sorted(M3.ribbon())
            [1, 2, 3, 4]
        """
        return self._ribbon

    def anchor(self):
        r"""
        Return the anchor of the mobile poset.

        EXAMPLES::

            sage: from sage.combinat.posets.mobile import MobilePoset
            sage: M2 = MobilePoset(Poset([[0,1,2,3,4,5,6,7,8],
            ....:          [(1,0),(3,0),(2,1),(2,3),(4,3),(5,4),(7,4),(7,8)]]))
            sage: M2.anchor()
            (4, 3)
            sage: M3 = MobilePoset(Posets.RibbonPoset(5, [1,2]))
            sage: M3.anchor() is None
            True
        """
        return self._anchor
